from functools import cache, wraps
from inspect import getfullargspec
from typing import Literal

import numpy as np
from scipy.interpolate import make_interp_spline

from pylegs.config import configs
from pylegs.io import read_table


def pseudo_vectorial(*vec_args):
  def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      args_name = getfullargspec(func)[0]
      args_dict = {**dict(zip(args_name, args)), **kwargs}
                
      is_vector = False
      vec_len = -1
      for k in vec_args:
        is_vector = is_vector or not np.isscalar(args_dict[k])
        if is_vector:
          vec_len = len(args_dict[k])
      
      if is_vector:
        vec_iter = ({k: args_dict[k][i] for k in vec_args} for i in range(vec_len))
        scalar_args = {k: v for k, v in args_dict.items() if k not in vec_args}
        return np.asarray([
          func(**scalar_args, **_vector_args)
          for _vector_args in vec_iter
        ])
      return func(*args, **kwargs)
    return wrapper
  return decorator


def compute_ellipse_bb(x, y, major, minor, angle_deg):
  """
  Compute tight ellipse bounding box.
  
  see https://stackoverflow.com/questions/87734/how-do-you-calculate-the-axis-aligned-bounding-box-of-an-ellipse#88020
  """
  eps = 1e-10
  angle_rad = np.radians(angle_deg)
  angle_rad_eps = np.radians(angle_deg + eps)
  
  t = np.arctan(-minor / 2 * np.tan(angle_rad_eps) / (major / 2))
  [min_x, max_x] = np.sort([
    x + major / 2 * np.cos(t) * np.cos(angle_rad) - 
    minor / 2 * np.sin(t) * np.sin(angle_rad) 
    for t in (t, t + np.pi)
  ], axis=0)
  
  t = np.arctan(minor / 2 * 1. / np.tan(angle_rad_eps) / (major / 2))
  [min_y, max_y] = np.sort([
    y + minor / 2 * np.sin(t) * np.cos(angle_rad) +
    major / 2 * np.cos(t) * np.sin(angle_rad) 
    for t in (t, t + np.pi)
  ], axis=0)
  
  return min_x, min_y, max_x, max_y



# @pseudo_vectorial('shape_e1', 'shape_e2')
def compute_ellipse_params(shape_e1: float, shape_e2: float, max_e: float = None):
  eps = 1e-10
  e = np.sqrt(shape_e1 ** 2 + shape_e2 ** 2)
  if max_e is not None:
    e = np.minimum(e, max_e)
  b = 1 - e
  a = 1 + e
  angle = 180 - np.rad2deg(np.arctan2(shape_e2 + eps, shape_e1 + eps) / 2)
  mask = np.greater(b, a)
  if np.isscalar(mask):
    a, b = b, a
    angle += 180
  else:
    temp = a[mask]
    a[mask] = b[mask]
    b[mask] = temp
    angle[mask] += 180
  return a, b, angle



@cache
def _get_correction_factor_df():
  return read_table(configs.CORRECTION_FACTOR_PATH)



@cache
def _get_correction_factor_model(
  kind: Literal['circle', 'ellipse'], 
  interp: Literal['step', 'linear']
):
  df = _get_correction_factor_df()
  if kind.lower() == 'ellipse' and interp.lower() == 'step':
    return make_interp_spline(df.mag_r.values, df.cf_ellip_step.values, 0)
  elif kind.lower() == 'ellipse' and  interp.lower() == 'linear':
    return make_interp_spline(df.mag_r.values + 0.5, df.cf_ellip_linear.values, 1)
  elif kind.lower() == 'circle' and interp.lower() == 'step':
    return make_interp_spline(df.mag_r.values, df.cf_circ_step.values, 0)
  elif kind.lower() == 'circle' and interp.lower() == 'linear':
    return make_interp_spline(df.mag_r.values + 0.5, df.cf_circ_linear.values, 1)
  


def correction_factor_ellipse(
  mag_r: float, 
  interp: Literal['step', 'linear'] = 'linear'
):
  if interp == 'linear':
    return _get_correction_factor_model('ellipse', 'linear')(mag_r)
  return _get_correction_factor_model('ellipse', 'step')(mag_r)



def correction_factor_circle(
  mag_r: float, 
  interp: Literal['step', 'linear'] = 'linear'
):
  if interp == 'linear':
    return _get_correction_factor_model('circle', 'linear')(mag_r)
  return _get_correction_factor_model('circle', 'step')(mag_r)



# @pseudo_vectorial('shape_e1', 'shape_e2', 'mag_r')
def compute_fov_ellip(shape_e1, shape_e2, mag_r, interp):
  a, b, angle = compute_ellipse_params(shape_e1, shape_e2, 0.6)
  x0, y0, x1, y1 = compute_ellipse_bb(0, 0, b, a, angle)
  cf = correction_factor_ellipse(mag_r, interp=interp)
  width = np.abs(x1 - x0) * 2
  height = np.abs(y1 - y0) * 2
  return np.maximum(width, height) * cf * 60



# @pseudo_vectorial('shape_e1', 'shape_e2', 'mag_r')
def compute_pixscale_ellip(shape_e1, shape_e2, size, mag_r, interp):
  fov = compute_fov_ellip(shape_e1=shape_e1, shape_e2=shape_e2, mag_r=mag_r, interp=interp)
  return fov / size



# @pseudo_vectorial('shape_r', 'mag_r')
def compute_fov_circ(shape_r, mag_r, interp):
  cf = correction_factor_circle(mag_r, interp=interp)
  return 2 * shape_r * cf



# @pseudo_vectorial('shape_r', 'mag_r')
def compute_pixscale_circle(shape_r, size, mag_r, interp):
  fov = compute_fov_circ(shape_r=shape_r, mag_r=mag_r, interp=interp)
  return fov / size


  


if __name__ == '__main__':
  df = read_table('/home/natan/repos/legacy-stamps/samples/sample_10-11_300.csv')
  x = compute_pixscale_ellip(df.shape_e1.values, df.shape_e2.values, 300, df.mag_r.values, 'linear')
  y = df.pixscale_ellip_linear.values
  print(np.allclose(x, y))
  # print(compute_ellipse_bb(0, 0, 10, 15, 45))
  # print(compute_ellipse_bb(0, 0, np.asarray([10, 20]), np.asarray([15, 30]), np.asarray([45, -45])))


"""
High-level interface to Legacy Survey data
"""


from io import BytesIO
from pathlib import Path
from typing import List, Literal, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import requests
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.units import Quantity
from PIL import Image

from pylegs.config import configs
from pylegs.io import (PathLike, PathOrFile, _is_folder, _prepare_path,
                       compress_fits_image, download_file,
                       parallel_function_executor)
from pylegs.pixscale import compute_pixscale_circle, compute_pixscale_ellip
from pylegs.utils import iauname, iauname_path


def _coord_to_deg(
  coord: int | float | Quantity | SkyCoord = None,
  ra: int | float | Quantity = None,
  dec: int | float | Quantity = None,
) -> Tuple[float, float]:
  if coord and isinstance(coord, SkyCoord):
    return coord.ra.to(u.deg).value, coord.dec.to(u.deg).value
  
  if isinstance(ra, (float, int)):
    new_ra = ra
  if isinstance(ra, Quantity):
    new_ra = ra.to(u.deg).value
  
  if isinstance(dec, (float, int)):
    new_dec = dec
  if isinstance(dec, Quantity):
    new_dec = dec.to(u.deg).value
  
  return new_ra, new_dec


def cutout(
  coord: SkyCoord = None,
  ra: int | float | Quantity = None,
  dec: int | float | Quantity = None,
  save_path: PathOrFile = None,
  overwrite: bool = False,
  width: int = None,
  height: int = None,
  size: int = None,
  fmt: Literal['jpg', 'fits'] = 'jpg',
  bands: str = None,
  pixscale: float = 0.3,
  layer: str = 'ls-dr10',
  compress_fits: bool = False,
  compress_type: str = 'HCOMPRESS_1',
  hcomp_scale: int = 3,
  quantize_level: int = 10,
  quantize_method: int = -1,
  http_client: requests.Session = None,
  dtype: Literal['bytes', 'numpy', 'pil', 'astropy'] = 'bytes'
) -> bytes | np.ndarray | Image.Image:
  """
  Downloads a single Legacy Survey object RGB stamp defined by RA and DEC.

  Parameters
  ----------
  ra: float
    Right ascension of the object.
  dec: float
    Declination of the object.
  save_path: pathlib.Path, optional
    Path where downloaded file will be stored.
  base_path: str, pathlib.Path, optional
    The path that will be appended at beggining of every paths if ``save_path``
    is ``None``.
  """
  ra, dec = _coord_to_deg(coord=coord, ra=ra, dec=dec)
  
  if fmt[0] == '.': fmt = fmt[1:]
    
  if fmt == 'jpg':
    url = configs.CUTOUT_RGB_BASE_URL
  else:
    url = configs.CUTOUT_FITS_BASE_URL
    
  if size is not None:
    width, height = size, size
  else:
    if width is None or height is None:
      width, height = 300, 300
      
  _prepare_path(save_path)

  if _is_folder(save_path):
    save_path = iauname_path(
      iaunames=iauname(ra=ra, dec=dec),
      prefix=save_path,
      suffix=f'.{fmt}'
    )

  query_params = {
    'ra': ra,
    'dec': dec,
    'width': width,
    'height': height,
    'pixscale': pixscale,
    'layer': layer
  }
  
  if bands is not None:
    query_params = {**query_params, 'bands': bands}

  image_bytes = download_file(
    url=url,
    query=query_params,
    save_path=save_path,
    http_client=http_client or configs.HTTP_CLIENT,
    overwrite=overwrite,
  )

  if compress_fits:
    buffer = BytesIO(image_bytes)
    buffer.seek(0)
    compress_fits_image(
      file=buffer,
      compress_type=compress_type,
      hcomp_scale=hcomp_scale,
      quantize_level=quantize_level,
      quantize_method=quantize_method,
      ext=0,
      save_path=save_path,
      overwrite=overwrite,
    )
    
  if dtype.lower() == 'bytes':
    return image_bytes
  if dtype.lower() == 'pil':
    return Image.open(BytesIO(image_bytes))
  if dtype.lower() == 'numpy':
    return np.asarray(Image.open(BytesIO(image_bytes)))
  if dtype.lower() == 'astropy':
    return fits.open(BytesIO(image_bytes))



  
def batch_cutout(
  coords: SkyCoord = None,
  ra: Sequence[int | float | Quantity | SkyCoord] | SkyCoord | None = None,
  dec: Sequence[int | float | Quantity | SkyCoord] | SkyCoord | None = None,
  save_path: Sequence[PathLike] | PathLike | None = None,
  workers: int = 1,
  overwrite: bool = False,
  width: int | None = None,
  height: int | None = None,
  size: int | None = None,
  fmt: Literal['jpg', 'fits'] = 'jpg',
  bands: str | None = None,
  layer: str = 'ls-dr10',
  pixscale: Sequence[float | int] | float | Literal['auto'] = 0.3,
  shape_e1: Sequence[float] | None = None,
  shape_e2: Sequence[float] | None = None,
  shape_r: Sequence[float] | None = None,
  mag_r: Sequence[float] | None = None,
  pixscale_interpolation: Literal['step', 'linear'] = 'linear',
  compress_fits: bool = False,
  compress_type: str = 'HCOMPRESS_1',
  hcomp_scale: int = 3,
  quantize_level: int = 10,
  quantize_method: int = -1,
  http_client: requests.Session | None = None,
) -> Tuple[List[Path], List[Path]]:
  """
  Downloads a list of objects defined by RA and DEC coordinates.

  The ``ra``, ``dec`` and ``save_path`` lists are mandatory and
  must have same length.

  Parameters
  ----------
  ra: List[float]
    The list of RA coordinates of the desired objects.
  dec: List[float]
    The list of DEC coordinates of the desired objects.
  save_path: List[Path], optional
    The list of path where files should be saved.
  base_path: str, Path, optional
    The path that will be appended at beggining of every paths if ``save_path``
    is ``None``.
  """
  ra, dec = _coord_to_deg(coord=coords, ra=ra, dec=dec)
  
  if _is_folder(save_path) or not Path(save_path).exists():
    save_path = iauname_path(
      iaunames=iauname(ra=ra, dec=dec),
      prefix=Path(save_path),
      suffix=f'.{fmt}'
    )

  params = [
    {
      'ra': _ra,
      'dec': _dec,
      'save_path': _save_path,
      'overwrite': overwrite,
      'width': width,
      'height': height,
      'size': size,
      'bands': bands,
      'layer': layer,
      'fmt': fmt,
      'compress_fits': compress_fits,
      'compress_type': compress_type,
      'hcomp_scale': hcomp_scale,
      'quantize_level': quantize_level,
      'quantize_method': quantize_method,
      'http_client': http_client,
    }
    for _ra, _dec, _save_path in zip(ra, dec, save_path)
  ]
  
  if pixscale == 'auto':
    if mag_r is None:
      raise ValueError('mag_r must be set to compute pixscale')
      
    if shape_r is not None:
      pixscale = compute_pixscale_circle(
        shape_r=shape_r, 
        size=size, 
        mag_r=mag_r, 
        interp=pixscale_interpolation,
      )
    elif shape_e1 is not None and shape_e2 is not None:
      pixscale = compute_pixscale_ellip(
        shape_e1=shape_e1, 
        shape_e2=shape_e2, 
        size=size, 
        mag_r=mag_r, 
        interp=pixscale_interpolation,
      )
    else:
      raise ValueError('set shape_e1 and shape_e2 for ellipse fit or shape_r for radial fit')
  
  if np.isscalar(pixscale):
    params = [{**_param, 'pixscale': pixscale} for _param in params]
  else:
    params = [{**_param, 'pixscale': _pixscale} for _param, _pixscale in zip(params, pixscale)]

  parallel_function_executor(
    cutout,
    params=params,
    workers=workers,
    unit='file',
    ignore_error=False,
  )

  success = [p for p in save_path if p.exists()]
  error = [p for p in save_path if not p.exists()]
  return success, error
    


if __name__ == '__main__':
  # ls = LegacyService(workers=6)
  # ls.batch_download_legacy_rgb(
  #   ra=[185.1458 + dx/2 for dx in range(20)],
  #   dec=[12.8624 + dy/2 for dy in range(20)],
  #   save_path=[Path(f'test/{i}.jpg') for i in range(20)]
  # )
  df = pd.DataFrame(({'ra':[184.4924, 184.4922], 'dec': [7.2737, 7.1862]}))
  cutout(ra=184.4924, dec=7.2737)
  # ls_crossmatch(df, ['mag_r', 'mag_g', 'mag_i', 'mag_z'], 3, 'test.csv')


from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Sequence

from astropy import units as u

from pylegs.cutout import batch_cutout
from pylegs.io import read_table
from pylegs.tap import DEFAULT_CATALOG_COLUMNS, download_catalog
from pylegs.utils import guess_coords_columns


def _is_numeric(value: str):
  try:
    float(value)
    return True
  except ValueError:
    return False


def handle_dlcat(args: Namespace):
  exclude = {'PSF', 'REX', 'DEV', 'EXP', 'SER'} - set([s.upper() for s in args.types])
  download_catalog(
    save_path=args.output,
    columns=args.cols,
    ra_min=args.ra[0],
    ra_max=args.ra[1],
    delta_ra=args.delta,
    table=args.table,
    exclude_types=exclude,
    magr_min=args.r[0],
    magr_max=args.r[1],
    dec_min=args.dec[0],
    dec_max=args.dec[1],
    brick_primary=not args.noprimary,
    overwrite=args.overwrite,
    workers=args.workers,
    tmp_folder=args.tmp,
    overwrite_tmp=args.otmp,
  )
  
def handle_cutout(args: Namespace):
  cols = []
  if args.ra is not None:
    cols.append(args.ra)
  if args.dec is not None:
    cols.append(args.dec)
  if not _is_numeric(args.pixscale) and args.pixscale != 'auto':
    cols.append(args.pixscale)
  
  df = read_table(args.cat, columns=cols or None)
  if args.limit is None:
    df = df.iloc[args.offset:]
  else:
    df = df.iloc[args.offset:args.limit]
  
  ra_col, dec_col = guess_coords_columns(df, args.ra, args.dec)
  ra = df[ra_col].values * u.deg
  dec = df[dec_col].values * u.deg
  shape_e1 = None
  shape_e2 = None
  shape_r = None
  mag_r = None
  if args.pixscale == 'auto':
    pixscale = 'auto'
    mag_r = df[args.magr].values
    if args.shape1 is not None and args.shape2 is not None:
      shape_e1 = df[args.shape1].values
      shape_e2 = df[args.shape2].values
    elif args.shaper is not None:
      shape_r = df[args.shaper].values
  elif _is_numeric(args.pixscale):
    pixscale = float(args.pixscale)
  else:
    pixscale = df[args.pixscale].values
  
  save_path = Path(args.output)
  if args.fnames is not None:
    save_path = [save_path / f'{f}.{args.format}' for f in df[args.fnames].values]
  
  batch_cutout(
    ra=ra,
    dec=dec,
    pixscale=pixscale,
    save_path=save_path,
    workers=args.workers,
    overwrite=args.overwrite,
    width=args.width,
    height=args.height,
    size=args.size,
    shape_e1=shape_e1,
    shape_e2=shape_e2,
    shape_r=shape_r,
    mag_r=mag_r,
    fmt=args.format,
    bands=args.bands,
    layer=args.layer,
    compress_fits=args.compress,
  )
  

def handle_axmatch(args: Namespace):
  from pylegs.archive import CrossMatcher
  cm = CrossMatcher.from_table(
    table=args.input, 
    radius=args.radius, 
    ra_col=args.ra, 
    dec_col=args.dec, 
    fmt=args.fmt,
    unit=args.unit,
  )
  cm.match(
    output_path=args.output,
    bricks_dir=args.cache,
    columns=args.columns,
    overwrite=args.overwrite,
    include_brickname=args.brickname,
    include_dr=args.dr,
  )
  
  
  
def handle_txmatch(args: Namespace):
  from pylegs.tap import ls_crossmatch
  ls_crossmatch(
    table=args.input,
    columns=args.columns,
    radius=args.radius,
    save_path=args.output,
    ra_col=args.ra,
    dec_col=args.dec,
    overwrite=args.overwrite,
    workers=args.workers,
  )



def entrypoint():
  parser = ArgumentParser(
    prog='pylegs', 
    description='Python client for accessing Legacy Survey data'
  )
  
  subparser = parser.add_subparsers(dest='subprog')
  dlcat = subparser.add_parser(
    'dlcat', help='Tool for download Legacy Survey catalog based on square area'
  )
  dlcat.add_argument(
    'output', action='store', type=str, help='Path of the output table'
  )
  dlcat.add_argument(
    '--ra', action='store', nargs=2, type=float, default=[0, 360], 
    help='Two values specifying the RA range in degrees. Default: 0 360'
  )
  dlcat.add_argument(
    '--delta', action='store', default='10 arcmin',
    help=('The value of the angle interval that the ra axis will be divided. '
          'The unit must be specified. E.g: 1.2 deg, 20 arcmin, 14 arcsec. '
          'Default: 10 arcmin')
  )
  dlcat.add_argument(
    '--dec', action='store', nargs=2, type=float, default=[-90, 90], 
    help='Two values specifying the DEC range in degrees. Default: -90 90'
  )
  dlcat.add_argument(
    '--cols', action='store', nargs='+', default=None,
    help=f'Column names to retrieve. Default: {" ".join(DEFAULT_CATALOG_COLUMNS)}'
  )
  dlcat.add_argument(
    '--table', action='store', default='ls_dr10.tractor',
    help='Fully qualified table name following the format: <schema_name>.<table_name>. Default: ls_dr10.tractor'
  )
  dlcat.add_argument(
    '--types', action='store', nargs='+', default=['REX', 'DEV', 'EXP', 'SER'],
    choices=['PSF', 'REX', 'DEV', 'EXP', 'SER'],
    help=('Morphological types to allow in final table.'
          'By default, only PSF type (stars) are removed. '
          'Possible values: PSF, REX, DEV, EXP, SER. Default: REX DEV EXP SER')
  )
  dlcat.add_argument(
    '--r', action='store', type=float, nargs=2, default=[10, 21],
    help='The magnitude range in band r. Default: 10 21'
  )
  dlcat.add_argument(
    '--noprimary', action='store_true', 
    help='Include no primary objects. By default, only primary objects are included.'
  )
  dlcat.add_argument(
    '--overwrite', action='store_true', 
    help='Use this flag to overwrite the destinaton path if the file exists'
  )
  dlcat.add_argument(
    '--workers', action='store', type=int, default=7,
    help='Number of parallel queries that will be sent to server. Default: 7'
  )
  dlcat.add_argument(
    '--tmp', action='store', default=None,
    help='Temp folder to store partial tables. Default: none'
  )
  dlcat.add_argument(
    '--otmp', action='store_true',
    help='Set this flag to overwrite partial tables in temp folder, if exists'
  )
  
  
  
  cutout = subparser.add_parser('cutout', help='Legacy stamps cutout')
  cutout.add_argument(
    'output', action='store',
    help='The folder path to save the figures'
  )
  cutout.add_argument(
    '--cat', '-c', action='store', default=None,
    help='The path of the catalog to be downloaded. Default: None'
  )
  cutout.add_argument(
    '--ra', action='store', default=None,
    help='The name of the RA column. Default: auto, the package will try to infer',
  )
  cutout.add_argument(
    '--dec', action='store', default=None,
    help='The name of the DEC column. Default: auto, the package will try to infer',
  )
  cutout.add_argument(
    '--fnames', '-f', action='store', default=None,
    help=(
      'The name of the column to be used as filename for each stamp.'
      'If not specified, this package will compute the IAUNAME for each '
      'object based on position RA and DEC and will use the IAUNAME as filename. '
      'Default: None'
    )
  )
  cutout.add_argument(
    '--pixscale', '-p', action='store', default=0.4,
    help=(
      'A float indicating a static pixscale for all stamps in arcsec/pixel, '
      'a string with the name of the pixscale column in the catalog, '
      'or "auto" to compute pixscale automatically. For the last option, '
      'mag_r, shape_e1 and shape_e2 must be set to compute pixscale using '
      'elliptical fit or mag_r and shape_r to use radial fit. Default: 0.4'
    ),
  )
  cutout.add_argument(
    '--limit', '-l', action='store', type=int, default=None,
    help=(
     'The row limit to download, starting with 0. '
     'Default: None (download the full catalog)'
    )
  )
  cutout.add_argument(
    '--offset', '-o', action='store', type=int, default=0,
    help=(
      'The index of the row to start the download, starting with 0. '
      'Default: 0 (the first row)'
    )
  )
  cutout.add_argument(
    '--format', action='store', default='jpg', choices=['jpg', 'fits'],
    help='The image format. Default: jpg'
  )
  cutout.add_argument(
    '--bands', action='store', default=None,
    help=(
      'A string representating the bands (griz) to dowload. '
      'Default: None (all available bands - griz)'
    )
  )
  cutout.add_argument(
    '--layer', action='store', default=None,
    help='The image layer. More details in Legacy documentation. Default: ls10-griz'
  )
  cutout.add_argument(
    '--width', action='store', type=int, default=256,
    help=(
      'The width of the image in pixels. Default: 256'
    )
  )
  cutout.add_argument(
    '--height', action='store', type=int, default=256,
    help=(
      'The height of the image in pixels. Default: 256'
    )
  )
  cutout.add_argument(
    '--size', '-s', action='store', type=int, default=None,
    help=(
      'The size of the image in pixels (square image). The arguments --width '
      'and --height are ignored if this argument is specified. Default: 256'
    )
  )
  cutout.add_argument(
    '--shape1', action='store', default=None,
    help='Name of the shape_e1 column'
  )
  cutout.add_argument(
    '--shape2', action='store', default=None,
    help='Name of the shape_e2 column'
  )
  cutout.add_argument(
    '--shaper', action='store', default=None,
    help='Name of the shape_r column'
  )
  cutout.add_argument(
    '--magr', action='store', default=None,
    help='Name of the mag_r column'
  )
  cutout.add_argument(
    '--compress', action='store_true',
    help='Compress fits stamp'
  )
  cutout.add_argument(
    '--overwrite', action='store_true', 
    help='Overwrite stamps'
  )
  cutout.add_argument(
    '--workers', action='store', type=int, default=6,
    help=(
      'The number of workers to spawn, meaning the number of parallel downloads. '
      'Default: 6'
    )
  )
  
  
  axmatch = subparser.add_parser('axmatch', help='Crossmatch Archive Server')
  axmatch.add_argument(
    'input', action='store', 
    help='Input table to crossmatch against Legacy Archive Server'
  )
  axmatch.add_argument(
    'output', action='store',
    help='Output table path'
  )
  axmatch.add_argument(
    '--radius', action='store', default='1 arcsec',
    help=(
      'Crossmatch radius, the unit must be specified. e.g. --radius 1.5 arcsec.'
      ' Default: 1 arcsec'
    )
  )
  axmatch.add_argument(
    '--cache', action='store', default='legacy_bricks',
    help=(
      'The cache dir. This folder will store the partial tables.' 
      ' Default: "legacy_bricks"'
    )
  )
  axmatch.add_argument(
    '--columns', action='store', nargs='+',
    help=(
      f'The columns to get from service. Default: {" ".join(DEFAULT_CATALOG_COLUMNS)}'
    )
  )
  axmatch.add_argument(
    '--ra', action='store', default=None,
    help='The name of the RA column. Default: auto, the package will try to infer',
  )
  axmatch.add_argument(
    '--dec', action='store', default=None,
    help='The name of the DEC column. Default: auto, the package will try to infer',
  )
  axmatch.add_argument(
    '--ifmt', action='store', default=None,
    help=(
      'Input table format, e.g. csv, parquet, fits. Default: auto, the package '
      'will try to infer from filename'
    )
  )
  axmatch.add_argument(
    '--unit', action='store', default='deg',
    help=(
      'Unit of the coordinates columns (RA and DEC) in inpu table. '
      'Default: deg'
    )
  )
  axmatch.add_argument(
    '--overwrite', action='store_true',
    help='Overwrites the output table'
  )
  axmatch.add_argument(
    '--brickname', action='store_true',
    help='Add brick name column in output table'
  )
  axmatch.add_argument(
    '--dr', action='store_true',
    help='Add data release column in output table'
  )
  
  
  
  txmatch = subparser.add_parser('txmatch', help='Crossmatch TAP Server')
  txmatch.add_argument(
    'input', action='store', 
    help='Input table to crossmatch against Legacy TAP Server'
  )
  txmatch.add_argument(
    'output', action='store',
    help='Output table path'
  )
  txmatch.add_argument(
    '--radius', action='store', default='1 arcsec',
    help=(
      'Crossmatch radius, the unit must be specified. e.g. --radius 1.5 arcsec.'
      ' Default: 1 arcsec'
    )
  )
  txmatch.add_argument(
    '--columns', action='store', nargs='+',
    help=(
      f'The columns to get from service. Default: {" ".join(DEFAULT_CATALOG_COLUMNS)}'
    )
  )
  txmatch.add_argument(
    '--ra', action='store', default=None,
    help='The name of the RA column. Default: auto, the package will try to infer',
  )
  txmatch.add_argument(
    '--dec', action='store', default=None,
    help='The name of the DEC column. Default: auto, the package will try to infer',
  )
  txmatch.add_argument(
    '--overwrite', action='store_true',
    help='Overwrites the output table'
  )
  txmatch.add_argument(
    '--workers', action='store', type=int, default=8,
    help='Number of parallel queries that will be sent to server. Default: 8'
  )
  
  
  
  handler_map = {
    'dlcat': handle_dlcat,
    'cutout': handle_cutout,
    'axmatch': handle_axmatch,
    'txmatch': handle_txmatch,
  }
  args = parser.parse_args()
  handler = handler_map.get(args.subprog)
  if handler:
    handler(args)
  else:
    parser.print_help()
  


if __name__ == '__main__':
  entrypoint()